from .cloud_model import CloudModel 
from scipy import interpolate
from scipy import optimize
import numpy as np 
from cmath import sqrt
from pathlib import Path

class RelScalar(CloudModel):
    """
    Relativistic (alpha~1) model for scalar bosons

    All in inputs are in units where black hole mass=1
    alpha is boson mass (times black hole mass)
    abh is black hole dimensionless spin
    Mcloud is boson cloud mass (as fraction of black hole mass) 
    m is azimuthal number of cloud
    """
    def __init__(self):
        """Caching tabulated and fit data"""

        """ 
        m=1 modes: 
        """
        m1_data = np.load(Path(__file__).parent.joinpath('data/m1_sc_mds.npz'))
        m1_wr = m1_data['wr'].flatten()
        m1_wi = m1_data['wi'].flatten()
        m1_a = m1_data['a'].flatten()
        m1_y = m1_data['y'].flatten()
        m1_dwr = m1_data['dwr'].flatten()
        self._f1wr = interpolate.LinearNDInterpolator(list(zip(m1_y,m1_a)),m1_wr)
        self._f1dwr = interpolate.LinearNDInterpolator(list(zip(m1_y,m1_a)),m1_dwr)
        self._f1wi = interpolate.LinearNDInterpolator(list(zip(m1_y,m1_a)),m1_wi)

        """ 
        m=2 modes: 
        """
        m2_data = np.load(Path(__file__).parent.joinpath('data/m2_sc_mds.npz'))
        m2_wr = m2_data['wr'].flatten()
        m2_wi = m2_data['wi'].flatten()
        m2_a = m2_data['a'].flatten()
        m2_y = m2_data['y'].flatten()
        m2_dwr = m2_data['dwr'].flatten()
        self._f2wr = interpolate.LinearNDInterpolator(list(zip(m2_y,m2_a)),m2_wr)
        self._f2dwr = interpolate.LinearNDInterpolator(list(zip(m2_y,m2_a)),m2_dwr)
        self._f2wi = interpolate.LinearNDInterpolator(list(zip(m2_y,m2_a)),m2_wi)

        """
        Fit coefficients
        """
        fit_data = np.load(Path(__file__).parent.joinpath('data/sc_fits.npz'))
        self._amat1 = fit_data['amat1']
        self._bmat1 = fit_data['bmat1']
        self._cmat1 = fit_data['cmat1']
        self._amat2 = fit_data['amat2']
        self._bmat2 = fit_data['bmat2']
        self._cmat2 = fit_data['cmat2']

        """
        Radiation data
        """
        sat_flux_data = np.load(Path(__file__).parent.joinpath('data/sc_sat_gw.npz'))
        m1_flux = sat_flux_data['m1_flux']
        m1_mu = sat_flux_data['m1_mu']
        m1_Z2r = sat_flux_data['m1_z2r']
        m1_Z2i = sat_flux_data['m1_z2i']
        m1_Z3r = sat_flux_data['m1_z3r']
        m1_Z3i = sat_flux_data['m1_z3i']
        m2_flux = sat_flux_data['m2_flux']
        m2_mu = sat_flux_data['m2_mu']
        m2_Z4r = sat_flux_data['m2_z4r']
        m2_Z4i = sat_flux_data['m2_z4i']
        m2_Z5r = sat_flux_data['m2_z5r']
        m2_Z5i = sat_flux_data['m2_z5i']

        self._pwm1 = interpolate.interp1d(m1_mu, m1_flux, kind='cubic')
        self._z2rm1 = interpolate.interp1d(m1_mu, m1_Z2r, kind='cubic')
        self._z2im1 = interpolate.interp1d(m1_mu, m1_Z2i, kind='cubic')
        self._z3rm1 = interpolate.interp1d(m1_mu, m1_Z3r, kind='cubic')
        self._z3im1 = interpolate.interp1d(m1_mu, m1_Z3i, kind='cubic')
        self._pwm2 = interpolate.interp1d(m2_mu, m2_flux, kind='cubic')
        self._z4rm2 = interpolate.interp1d(m2_mu, m2_Z4r, kind='cubic')
        self._z4im2 = interpolate.interp1d(m2_mu, m2_Z4i, kind='cubic')
        self._z5rm2 = interpolate.interp1d(m2_mu, m2_Z5r, kind='cubic')
        self._z5im2 = interpolate.interp1d(m2_mu, m2_Z5i, kind='cubic')

    def max_azi_num(self):
        """Maximum azimuthal number the model is defined for"""
        return 2
    def max_spin(self):
        """Maximum spin the model is defined for"""
        return 0.995 
    def omega_real(self, m, alpha, abh, Mcloud):
        """Returns real frequency of cloud oscillation"""
        yl, alphamax, Oh = self._y(m, alpha, abh)
        dwr = self._deltaomega(m, alpha, abh)
        if (m==1):
            if (alpha>=0.05 and abh>=self._aswitch(m)):
                wr = alpha*self._f1wr(yl, abh)
            else:
                # Ensures output fails sup. condition if alpha too large
                if (not(self._maxalphaInterp(m, alpha))):
                    return np.nan
                wr = alpha*(1.0-alpha**2/8.0-17.0*alpha**4/128.0+abh*alpha**5/12.0)
                wr += alpha*0.0028391854462700*alpha**5*(abh*np.sqrt(1-abh**2)-abh)
                for p in range(6, 9):
                    for q in range(0, 4):
                        wr += alpha**(p+1)*(1.0-abh**2)**(q/2.0)*self._amat1[p-6,q]
            return wr+Mcloud*dwr
        elif (m==2):
            if (alpha>=0.25 and abh>=self._aswitch(m)):
                wr = alpha*self._f2wr(yl, abh)
            else:
                # Ensures output fails sup. condition if alpha too large
                if (not(self._maxalphaInterp(m, alpha))):
                    return np.nan
                wr = alpha*(1.0-alpha**2/18.0-23.0*alpha**4/1080.0+4.0*abh*alpha**5/405.0)
                wr += alpha*0.0024403837275569847*alpha**5*(abh*np.sqrt(1-abh**2)-abh)
                for p in range(6, 9):
                    for q in range(0, 4):
                        wr += alpha**(p+1)*(1.0-abh**2)**(q/2.0)*self._amat2[p-6,q]
            return wr+Mcloud*dwr
        else:
            raise ValueError("Azimuthal index too large")
    def domegar_dmc(self, m, alpha, abh, Mcloud):
        """Returns derivative of real frequency of cloud w.r.t. cloud mass"""
        return self._deltaomega(m, alpha, abh)
    def omega_imag(self, m, alpha, abh):
        """Returns imaginary frequency, 
           i.e. growth rate of superradiant instability"""
        yl, alphamax, Oh = self._y(m, alpha, abh)
        wr = self.omega_real(m, alpha, abh, 0)
        if (m==1):
            if (alpha>=0.05 and abh>=self._aswitch(m)):
                return -np.exp(self._f1wi(yl, abh))*(wr-m*Oh)
            else:
                wi = 1.0
                glm = 1-abh**2+(abh*m-2.0*wr*(1.0+np.sqrt(1.0-abh**2)))**2
                for p in range(1, 4):
                    for q in range(0, 4):
                        wi += alpha**p*(abh**(q+1)*self._bmat1[q,p-1]+self._cmat1[q,p-1]*(1.0-abh**2)**(q/2.0))
                wi *= -2.0*(1.0+np.sqrt(1.0-abh**2))*glm*alpha**9*(wr-m*Oh)/48.0
                return wi
        elif (m==2):
            if (alpha>=0.25 and abh>=self._aswitch(m)):
                return -np.exp(self._f2wi(yl, abh))*(wr-m*Oh)
            else:
                wi = 1.0
                glm = 1-abh**2+(2.0*abh-2.0*wr*(1.0+np.sqrt(1.0-abh**2)))**2
                glm *= 4.0*(1-abh**2)+(2.0*abh-2.0*wr*(1.0+np.sqrt(1.0-abh**2)))**2
                wi += -1.1948426572069112e11*alpha**12+2.609027546773062e12*alpha**13
                for p in range(12, 23):
                    wi += self._cmat2[p-12]*alpha**p*(1.0-abh**2)**0.5
                    for q in range(0, 3):
                        wi += alpha**(p+2)*abh**q*self._bmat2[q,p-12]
                wi *= -2.0*(1.0+np.sqrt(1.0-abh**2))*glm*alpha**13*(wr-m*Oh)*4.0/885735.0
                return wi
        else:
            raise ValueError("Azimuthal index too large")
    def power_gw(self, m, alpha, abh):
        """Returns gravitational wave power, scaled to Mcloud=1 
        Using: omega_R(Mcloud=0), valid only at saturation"""
        if (m==1):
            if (alpha<0.2):
                a14 = 0.0109289473739731
                a15 = -0.0290105840870259
                return a14*alpha**14+a15*alpha**15
            else:
                return self._pwm1(alpha)
        elif (m==2):
            if (alpha<0.34):
                a18 = 6.46575425669374e-7
                a19 = -1.12205283686066e-6
                return a18*alpha**18+a19*alpha**19
            else:
                return self._pwm2(alpha)
        else:
            raise ValueError("Azimuthal index too large")
    def strain_sph_harm(self, m, alpha, abh):
        """Returns e^{iwt}R h^{\ell 2m} (-2-weighted spherical harmonic components)"""
        spsat = self._spinsat(m, alpha)
        wr = 2.0*self.omega_real(m, alpha, spsat, 0)
        if (m==1):
            if (alpha<0.2):
                z2abs = 2.0*np.pi*np.sqrt(wr**2*self.power_gw(m, alpha, spsat))
                return 2.0*np.array([z2abs,0])/(np.sqrt(2.0*np.pi)*wr**2)
            else:
                z2 = self._z2rm1(alpha)+1j*self._z2im1(alpha)
                z3 = self._z3rm1(alpha)+1j*self._z3im1(alpha)
                return -2.0*np.array([z2, z3])/(np.sqrt(2.0*np.pi)*wr**2)
        elif (m==2):
            if (alpha<0.34):
                z4abs = 2.0*np.pi*np.sqrt(wr**2*self.power_gw(m, alpha, spsat))
                return 2.0*np.array([z4abs,0])/(np.sqrt(2.0*np.pi)*wr**2)
            else:
                z4 = self._z4rm2(alpha)+1j*self._z4im2(alpha)
                z5 = self._z5rm2(alpha)+1j*self._z5im2(alpha)
                return -2.0*np.array([z4, z5])/(np.sqrt(2.0*np.pi)*wr**2)
        else:
            raise ValueError("Azimuthal index too large")
    def _aswitch(self, m):
        return 0.6
    def _maxalphaInterp(self, m, alpha):
        """Returns True if alpha is below the saturation-alpha for each m at abh=aswitch(m)"""       
        if (m==1):
            return alpha<0.18
        elif (m==2):
            return alpha<0.4
        else:
            raise ValueError("Azimuthal index too large") 
    def _y(self, m, alpha, abh, beta=0.9):
        """Returns utility parameter y, approximate maximal alpha, horizon frequency"""
        if (m==1):
            alpha0 = 0.05
        else:
            alpha0 = 0.25
        Oh = 0.5*abh/(1.0+np.sqrt(1.0-abh**2))
        temp = 9.0*m*(m+1)**2*Oh*beta**2+sqrt(81*m**2*(1+m)**4*Oh**2*beta**4-24*(1+m)**6*beta**6)
        alphamaxC = ((3**(1.0/3.0)+1j*3**(5.0/6.0))*temp**(2.0/3.0)-4.0*(-3.0)**(2.0/3.0)*(m+1)**2*beta**2)/(6.0*temp**(1.0/3.0)*beta)
        alphamax = alphamaxC.real
        yl = (alpha-alpha0)/(alphamax-alpha0)
        return yl, alphamax, Oh
    def _deltaomega(self,m,alpha,abh):
        """Returns the cloud mass-independent frequency shift due to self-gravity"""
        yl, alphamax, Oh = self._y(m, alpha, abh)
        #Factor of 2 below is to go from gravitational potential energy to frequency shift
        if (m==1):
            if (alpha>=0.05 and abh>=self._aswitch(m)):
                return -2.0*alpha*self._f1dwr(yl, abh)
            else:
                dwr = -93*alpha**3/1024.0
                dwr += -alpha*(-0.0040394234400504087576*alpha**3 + 0.67372467642478683914*alpha**4 - 3.5059737204226921747*alpha**5 + 5.2355803557173512530*alpha**6)
                return 2.0*dwr
        elif (m==2):
            if (alpha>=0.25 and abh>=self._aswitch(m)):
                return -2.0*alpha*self._f2dwr(yl, abh)
            else:
                dwr = -793*alpha**3/18432.0
                dwr += alpha*(-0.13855421027039729887*alpha**3 + 1.159105481278182692*alpha**4 - 3.1372867270252249305*alpha**5 + 2.8141844762025582938*alpha**6)
                return 2.0*dwr
        else:
            raise ValueError("Azimuthal index too large")
    def _alphasat(self, m, abh):
        """Bisection to find saturation point, returns alpha at saturation"""
        yh, amaxh, Ohh = self._y(m, 0, abh)
        yl, amaxl, Ohl = self._y(m, 0, abh, 1.1)
        def _sat(al):
            return self.omega_real(m, al, abh, 0)-m*0.5*abh/(1.0+np.sqrt(1.0-abh**2))
        return optimize.bisect(_sat, amaxl, amaxh)
    def _spinsat(self, m, al):
        """Bisection to find saturation point, returns spin at saturation"""
        def _sat(abh):
            return self.omega_real(m, al, abh, 0)-m*0.5*abh/(1.0+np.sqrt(1.0-abh**2))
        return optimize.bisect(_sat, self.max_spin(), 0)       



from .nonrel_cloud import NonrelScalar, NonrelVector
from .rel_vec_cloud import RelVector
from .rel_sca_cloud import RelScalar
from .matched_waveform import MatchedWaveform 
from .full_evo_waveform import FullEvoWaveform 

class UltralightBoson(object):
    def __init__(self, spin=1, model="relativistic"):
        if (not (spin==0 or spin==1)):
            raise ValueError("Spin value %d not supported" % spin)
        if (model=="non-relativistic"):
            if (spin==0):
                self._cloud_model = NonrelScalar() 
            elif (spin==1):
                self._cloud_model = NonrelVector()
        elif (model=="relativistic"):
            if (spin==0):
                self._cloud_model = RelScalar()
            elif (spin==1):
                self._cloud_model = RelVector()
        else:
            raise ValueError("Model %s not supported" % model)
    def make_waveform(self, Mbh, abh, mu, units="physical", evo_type="matched"):
        if (evo_type=="matched"): 
            return MatchedWaveform(mu, Mbh, abh, self._cloud_model, units=units)
        elif (evo_type=="full"):
            return FullEvoWaveform(mu, Mbh, abh, self._cloud_model, units=units)
        else:
            raise ValueError("evo_type %s not supported" % evo_type)



import unittest
import numpy as np
import superrad.ultralight_boson as ulb
import superrad.rel_vec_cloud as rvc
import superrad.rel_sca_cloud as rsc
import superrad.nonrel_cloud as nrc

class TestFrontEnd(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        """Setting up front ends of all waveform models"""
        self.alp = [0.001,0.01,0.1,0.5,1.0]
        self.spi = [0.001,0.1,0.5,0.8,0.99,0.995]
        self.bcwm = []
        model_type = ['relativistic','non-relativistic']
        model_spin = [0,1]
        self.nmodels = len(model_spin)*len(model_type)
        label = 0
        self.conv = {}
        for ty in model_type:
            for sp in model_spin:
                self.bcwm.append(ulb.UltralightBoson(spin=sp,model=ty))
                self.conv[label] = ty+', spin = '+str(sp)
                label += 1

    def tearDown(self):
        pass

    def test_power_consistency(self):
        """Testing GW power against backend quantity in natural units"""
        tol = 1e-1

        theta = np.linspace(0,np.pi,513)
        for i,mod in enumerate(self.bcwm):
            for al in self.alp:
                for sp in self.spi:
                    try:
                        wm = mod.make_waveform(1,sp,al,units="natural")
                        pw = wm.power_gw(0)
                        wr = 2.0*np.pi*wm.freq_gw(1e20)*wm.mass_bh_final()
                        hp,hx,delta = wm.strain_amp(0, theta)
                        pwstr = wr**2*np.trapz((hp**2+hx**2)*np.sin(theta),theta)*np.pi/(16.0*np.pi)
                        if (abs((pwstr-pw)/pw)>tol):
                            print(self.conv[i])
                            print("alpha = "+str(al))
                            print("spin = "+str(sp))
                            print("diff = "+str(abs((pwstr-pw)/pw)))
                            self.assertTrue(abs((pwstr-pw)/pw)<tol)
                    except ValueError:
                        pass
       


    def test_saturation_conditions(self):
        """
        Testing saturation output of front end in all models given 
        some alpha and spin to within tolerance
        """
        tol = 1e-3

        for i,mod in enumerate(self.bcwm):
            for al in self.alp:
                for sp in self.spi:
                    try:
                        wm = mod.make_waveform(1,sp,al,units="natural")
                        spf = wm.spin_bh_final()
                        mf = wm.mass_bh_final()
                        wr = wm.freq_gw(0)*np.pi*mf
                        mi = wm.azimuthal_num()
                        Oh = 0.5*spf/(1.0+np.sqrt(1.0-spf**2))
                        if (abs((mi*Oh-wr)/wr)>tol):
                            print(self.conv[i])
                            print("alpha = "+str(al))
                            print("spin = "+str(sp))
                            print("diff = "+str(abs((mi*Oh-wr)/wr)))
                            self.assertTrue(abs((mi*Oh-wr)/wr)<tol)
                    except ValueError:
                        pass

    def test_phase(self):
        """
        Test that the time derivative of the GW phase gives the same value as the frequency.
        """
        tol = 1e-9

        for i,mod in enumerate(self.bcwm):
            for al in self.alp:
                for sp in self.spi:
                    try:
                        wm = mod.make_waveform(1,sp,al,units="natural")
                        t = np.linspace(0, 2*wm.gw_time(), 2048)
                        phi = wm.phase_gw(t)
                        phi_dot = (phi[1:]-phi[:-1])/(t[1:]-t[:-1]) 
                        th = 0.5*(t[1:]+t[:-1]) 
                        f = wm.freq_gw(th)
                        rel_err = abs((f-phi_dot/(2.0*np.pi))/f)
                        #plt.plot(th, rel_err, ":")
                        #plt.show()
                        if (rel_err.max()>tol):
                            print(self.conv[i])
                            print("alpha = "+str(al))
                            print("spin = "+str(sp))
                            print("rel error = "+str(rel_err.max()))
                            self.assertTrue(rel_err.max()<tol)
                    except ValueError:
                        pass

if __name__ == '__main__':
    unittest.main()
                    



import sys
import numpy as np
from matplotlib import pyplot as plt 
import os
from pathlib import Path
current = os.path.dirname(os.path.realpath(__file__))
parent = os.path.dirname(current)
sys.path.append(parent)
from superrad.ultralight_boson import UltralightBoson


def plot_remnant_example():
    """Plot superradiant instability and GW timescale for example remnant"""
    models = ["relativistic", "non-relativistic"]
    mrk = ["-", ":"]
    Mbh0 = 80.0 #solar mass
    abh0 = 0.81
    mu_range = 10.0**np.linspace(-13.5,-12.0,256) # eV
    for n,model in enumerate(models):
        bc = UltralightBoson(spin=1, model=model) 
        tauI = np.zeros_like(mu_range) 
        tauGW = np.zeros_like(mu_range) 
        Mbhf = np.zeros_like(mu_range) 
        abhf = np.zeros_like(mu_range) 
        ms = np.zeros_like(mu_range) 
        fgw = np.zeros_like(mu_range) 
        for i,mu in enumerate(mu_range):
            try:
                wf = bc.make_waveform(Mbh0, abh0, mu, units="physical")
            except ValueError:
                break 
            tauI[i] = wf.efold_time()
            tauGW[i] = wf.gw_time()
            Mbhf[i] = wf.mass_bh_final()
            abhf[i] = wf.spin_bh_final()
            ms[i] = wf.azimuthal_num()
            fgw[i] = wf.freq_gw(0)
        plt.loglog(fgw[ms==1], (tauI)[ms==1]/(60.0)**2, "r"+mrk[n], label=model)
        plt.loglog(fgw[ms==1], (tauGW)[ms==1]/(60.0)**2, "k"+mrk[n])
        plt.loglog(fgw[ms==2], (tauI)[ms==2]/(60.0)**2, "g"+mrk[n])
        plt.loglog(fgw[ms==2], (tauGW)[ms==2]/(60.0)**2, "b"+mrk[n])
    plt.legend(loc="best")
    plt.xlabel(r"$f_{GW}$ (Hz)")
    plt.ylabel(r"$\tau$ (hours)")
    plt.show() 

def plot_freq_shift_example():
    """Compute the relative frequency shift after boson cloud has lost half of its mass"""
    models = ["relativistic", "non-relativistic"]
    mrk = ["-", ":"]
    Mbh0 = 1.0
    abh0 = 0.99
    mu_range = np.linspace(0.05, 1.1, 1024)
    for n,model in enumerate(models):
        bc = UltralightBoson(spin=0, model=model) 
        deltaf_f = np.zeros_like(mu_range) 
        for i,mu in enumerate(mu_range):
            try:
                wf = bc.make_waveform(Mbh0, abh0, mu, units="natural")
            except ValueError:
                break 
            tauGW = wf.gw_time()
            f = wf.freq_gw(np.array([0,tauGW]))
            deltaf_f[i] = (f[1]-f[0])/f[0]
        plt.semilogy(mu_range, deltaf_f, "k"+mrk[n], label=model)
    plt.legend(loc="best")
    plt.ylabel(r"$\Delta f /f$")
    plt.xlabel(r"$\mu M$")
    plt.show() 

def remnant_properties_example():
    """Print out some properties of an example remnant and plot evolution of GW
    signal"""
    bc = UltralightBoson(spin=1, model="relativistic") 
    Mbh = 20.8 #solar mass 
    abh = 0.7
    mu = 1.16e-12 # eV
    print("Initial black hole of %1.1f solar mass and dimensionless spin %1.2f" % (Mbh,abh))
    print("Ultralight boson of mass %1.2e eV" % mu)
    wf = bc.make_waveform(Mbh, abh, mu, units="physical")
    print("Cloud grows in  %1.1f hours" % (wf.cloud_growth_time()/(60.0*60.0)))
    print("At saturation cloud mass is %e solar mass" % wf.mass_cloud(0))
    print("At saturation cloud GW frequency is %1.1f Hz" % wf.freq_gw(0))
    print("GW timescale is %1.1f hours" % (wf.gw_time()/(60.0*60.0)))
    sec_hour = 3600.0
    thetaObs = np.pi/4
    dObs = 100.0 # Mpc
    t = np.linspace(0,24*sec_hour, 256)
    fgw = wf.freq_gw(t)
    fdotgw = wf.freqdot_gw(t)
    hp,hx,delta = wf.strain_amp(t, thetaObs, dObs=dObs)
    phi = wf.phase_gw(t)
    phi_dot = (phi[1:]-phi[:-1])/(t[1:]-t[:-1])
    th = 0.5*(t[1:]+t[:-1]) 

    fig, (ax1, ax2, ax3) = plt.subplots(3, 1, sharex=True)
    ax1.plot(t/sec_hour, hp, label=r"$h_+$")
    ax1.plot(t/sec_hour, hx, label=r"$h_x$")
    ax1.set_ylabel(r"$h$")
    ax1.legend(loc="best")
    ax2.plot(t/sec_hour, fgw, label=r"$f_{\rm GW}$")
    ax2.plot(th/sec_hour, phi_dot/(2.0*np.pi), ":", label=r"$\dot{\phi}/(2\pi)$")
    ax2.set_ylabel(r"$f_{\rm GW}$ (Hz)")
    ax2.legend(loc="best")
    ax3.plot(t/sec_hour, fdotgw)
    ax3.set_ylabel(r"$\dot{f}_{\rm GW}$ (Hz/s)")
    ax3.set_xlabel(r"$t$ (hours)")
    plt.show()

def gw_power_strain():
    """Check for consistency between GW power and strain"""
    models = ["relativistic", "non-relativistic"]
    mrk = ["-", ":"]
    Mbh0 = 1.0
    abh0 = 0.99
    mu_range = np.linspace(0.01, 1.1, 256)
    theta = np.linspace(0, np.pi, 256)
    for n,model in enumerate(models):
        bc = UltralightBoson(spin=1, model=model) 
        pgw = np.zeros_like(mu_range) 
        pgw_h = np.zeros_like(mu_range) 
        ms = np.zeros_like(mu_range) 
        for i,mu in enumerate(mu_range):
            try:
                wf = bc.make_waveform(Mbh0, abh0, mu, units="natural")
            except ValueError:
                break 
            pgw[i] = wf.power_gw(0)
            omegaGW = 2.0*np.pi*wf.freq_gw(1e20)*wf.mass_bh_final()
            hp,hx,delta = wf.strain_amp(0, theta)
            pgw_h[i] = omegaGW**2*np.trapz((hp**2+hx**2)*np.sin(theta),theta)*np.pi/(16.0*np.pi)
            ms[i] = wf.azimuthal_num()
        plt.semilogy(mu_range[ms==1], pgw[ms==1], "k"+mrk[n], label=model)
        plt.semilogy(mu_range[ms==1], pgw_h[ms==1], "r"+mrk[n], label=model+" from $h$")
        plt.semilogy(mu_range[ms==2], pgw[ms==2], "g"+mrk[n])
        plt.semilogy(mu_range[ms==2], pgw_h[ms==2], "b"+mrk[n])
    plt.ylabel(r"$P_{\rm GW}$")
    plt.xlabel(r"$\mu M$")
    plt.legend(loc="best")
    plt.show()
    
def gw_strain():
    """Compare GW strain"""
    models = ["relativistic", "non-relativistic"]
    mrk = ["-", ":"]
    Mbh0 = 1.0
    abh0 = 0.99
    mu_range = np.linspace(0.01, 1.1, 256)
    theta = 3*np.pi/4 
    for n,model in enumerate(models):
        bc = UltralightBoson(spin=0, model=model) 
        hp = np.zeros_like(mu_range) 
        hx = np.zeros_like(mu_range) 
        delta = np.zeros_like(mu_range) 
        ms = np.zeros_like(mu_range) 
        for i,mu in enumerate(mu_range):
            try:
                wf = bc.make_waveform(Mbh0, abh0, mu, units="natural")
            except ValueError:
                break 
            omegaGW = 2.0*np.pi*wf.freq_gw(0)
            hp[i],hx[i],delta[i] = wf.strain_amp(0, theta)
            ms[i] = wf.azimuthal_num()
        plt.semilogy(mu_range[ms==1], hp[ms==1], "k"+mrk[n], label=model+r" $h_+$")
        plt.semilogy(mu_range[ms==1], hx[ms==1], "r"+mrk[n], label=model+r" $h_x$")
        plt.semilogy(mu_range[ms==2], hp[ms==2], "g"+mrk[n])
        plt.semilogy(mu_range[ms==2], hx[ms==2], "b"+mrk[n])
    plt.ylabel(r"$h$")
    plt.xlabel(r"$\mu M$")
    plt.legend(loc="best")
    plt.show()

def comp_cloud_evo():
    """Compare a simple approximation for the cloud evolution that matches
    together the expotential growth phase of the cloud with the GW decay
    dominated phase (matched) to evolving cloud while accounting for both
    superradiant growth and GW decay (full).
    """
    bc = UltralightBoson(spin=1, model="relativistic") 
    Mbh = 20.8 #solar mass 
    abh = 0.7
    mu = 1.16e-12 # eV
    sec_hour = 3600.0
    thetaObs = np.pi/4
    dObs = 100.0 # Mpc
    t = np.linspace(-0.25*sec_hour,3*sec_hour, 2048)
    fig, (ax1, ax2, ax3) = plt.subplots(3, 1, sharex=True)
    for evo_type in ["matched", "full"]:
        wf = bc.make_waveform(Mbh, abh, mu, units="physical", evo_type=evo_type)
        Mc = wf.mass_cloud(t)
        fgw = wf.freq_gw(t)
        fdotgw = wf.freqdot_gw(t)
        mrk = "-"
        if (evo_type=="full"): mrk = "--"
        ax1.plot(t/sec_hour, Mc, mrk, label=evo_type)
        ax2.plot(t/sec_hour, fgw, mrk, label=evo_type)
        ax3.semilogy(t/sec_hour, abs(fdotgw), mrk, label=evo_type)
    ax1.set_ylabel(r"$M_c\ (M_{\odot})$")
    ax1.legend(loc="best")
    ax2.set_ylabel(r"$f_{\rm GW}$ (Hz)")
    ax2.legend(loc="best")
    ax3.set_ylabel(r"$|\dot{f}_{\rm GW}|$ (Hz/s)")
    ax3.set_xlabel(r"$t$ (hours)")
    plt.show()


if __name__ == "__main__":
    remnant_properties_example()
    plot_remnant_example()
    plot_freq_shift_example()
    gw_power_strain()
    gw_strain()
    comp_cloud_evo()




from abc import ABCMeta, abstractmethod

class BosonCloudWaveform(metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, mu, Mbh0, abh0, cloud_model, units="natural"):
        """
        Calculate some derived quantites using the specified CloudModel
        for the specified parameters:
        mu : ultralight boson mass
        Mbh0 : initial black hole mass (before cloud growth)
        abh0 : initial black hole dimensionless (before cloud growth)
        
        If units="physical" use the following units for input/output:

        mu : electronvolts
        Mass : solar mass
        time : seconds
        frequency : Hz
        Power : watts
        Distance : Mpc

        If units="natural" assume G=c=hbar=1

        If "+alpha" is appended to either "physical" or "natural," then units
        are the same as above, except the input mu is taken to be in units of 
        (hbar c^3)/(G*Mbh0), i.e.  mu is set to the dimensionless 
        "fine structure constant" alpha. 
        """
        pass
    @abstractmethod
    def azimuthal_num(self):
        """Azimuthal number of cloud (mass/gravitational waves have twice this)"""
        pass
    @abstractmethod
    def mass_bh_final(self):
        """Black hole mass at saturation"""
        pass
    @abstractmethod
    def spin_bh_final(self):
        """Black hole dimensionless spin at saturation"""
        pass
    @abstractmethod
    def efold_time(self):
        """Before saturation, e-folding time of boson cloud mass"""
        pass
    @abstractmethod
    def cloud_growth_time(self):
        """Time for cloud to grow from single boson to saturation"""
        pass
    @abstractmethod
    def mass_cloud(self, t):
        """Mass of boson cloud as  function of time"""
        pass
    @abstractmethod
    def power_gw(self, t):
        """Power (luminosity) of gravitational waves as a function of time"""
        pass
    @abstractmethod
    def gw_time(self):
        """Characteristic timescale of GW emission (Mc/P_GW) at saturation"""
        pass
    @abstractmethod
    def freq_gw(self, t):
        """Frequency of gravitational wave signal as a function of time"""
        pass
    @abstractmethod
    def freqdot_gw(self, t):
        """Time derivative of frequency of gravitational wave signal as a function of time"""
        pass
    @abstractmethod
    def strain_char(self, t, dObs=None): 
        """
        A characteristic strain value, defined to be: 
        h0:=(10 P_{GW})^{1/2}/(omega_{GW}*dObs).
        dObs is distance to source.

        In the non-relativistic limit (and for azimuthal_num=1), should have that:
        h_+ = h_0*(1+cos^2(theta))/2*cos(phi(t))
        h_x = h_0*cos(theta)*sin(phi(t))
        """
        pass
    @abstractmethod
    def strain_amp(self, t, thetaObs, dObs=None):
        """
        Returns magnitude of two polarizations of strain (h_+,h_x) as a
        function of time and inclination angle with respect to spin axis.  Also
        return delta, where delta is the extra phase difference between
        polarizations.  That is, observed strain is: 
        h_I = F^I_+(t)*h_+(t)cos(phi(t))+F^I_x*h_x(t)*sin(phi(t)+delta) 

        dObs is distance to source.
        """
        pass
    @abstractmethod
    def phase_gw(self, t):
        """
        Return the gravitational wave phase [phi(t) as defined above] as a
        function of time. Here we assume (take the approximation) that domega/dMc is a
        constant. By convention, the phase is zero at t=0.
        """
        pass
